
import numpy as np
cimport numpy as np
import logging
import os

from despasito.equations_of_state import constants

# For Numba, ckl_coef cannot be encapsulated
ckl_coef = np.array([[0.81096, 1.7888, -37.578, 92.284], [1.0205, -19.341, 151.26, -463.50],
                     [-1.9057, 22.845, -228.14, 973.92], [1.0885, -6.1962, 106.98, -677.64]])


def prefactor(l_r, l_a):
    r"""
    Calculations C, the Mie potential prefactor, defined in eq. 2
    
    Parameters
    ----------
    l_a : float
        Mie potential attractive exponent
    l_r : float
        Mie potential attractive exponent
    
    Returns
    -------
    C : float
        Mie potential prefactor
    """
    if np.any(l_r == 0.0):
        print("l_r should not be zero.")

    return (l_r / (l_r - l_a)) * (l_r / l_a)**(l_a / (l_r - l_a))

def calc_a1s(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r""" 
    Return a1s,kl(rho*Cmol2seg,l_kl) in K as defined in eq. 25.
    
    Used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hardsphere diameters for groups (k,l)

    Returns
    -------
    a1s : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Nrho x Ngroups x Ngroups
    """

    nbeads = len(dkl)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1,4):
        zetax_pow[:, i] = zetax_pow[:, i-1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etakl = np.zeros((len(rho), nbeads, nbeads), dtype=rho.dtype)

    for k in range(nbeads):
        for l in range(nbeads):
            tmp = np.dot(ckl_coef, np.array( (1.0, 1.0/l_kl[k, l], 1.0/l_kl[k, l]**2, 1.0/l_kl[k, l]**3), dtype=ckl_coef.dtype ))
            etakl[:, k, l] = np.dot( zetax_pow, tmp )
    a1s = - (1.0 - (etakl / 2.0)) / ((1.0 - etakl)**3) * 2.0 * np.pi * Cmol2seg * ((epsilonkl * (dkl**3 * constants.molecule_per_nm3**2)) / (l_kl - 3.0))

    output = np.transpose(np.transpose(a1s) * rho)

    return output

def calc_Bkl(rho, l_kl, Cmol2seg, dkl, epsilonkl, x0kl, zetax):
    r"""
    Return Bkl(rho*Cmol2seg,l_kl) in K as defined in eq. 20.
    
    Used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    l_kl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential exponents for k,l groups
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    x0kl : numpy.ndarray
        Matrix of sigmakl/dkl, ratio of Mie radius for groups scaled by hard sphere interaction (k,l)
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    Bkl : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is rho x l_kl.shape
    
    """

    rhos = Cmol2seg * rho

    # compute Ikl(l_kl), eq. 23
    Ikl = (1.0 - (x0kl**(3.0 - l_kl))) / (l_kl - 3.0)
    # compute Jkl(l_kl), eq. 24
    Jkl = (1.0 - ((x0kl**(4.0 - l_kl)) * (l_kl - 3.0)) + ((x0kl**(3.0 - l_kl)) * (l_kl - 4.0))) / ((l_kl - 3.0) * (l_kl - 4.0))

    tmp11 = rhos * (2.0 * np.pi)
    tmp12 = (dkl**3 * constants.molecule_per_nm3**2) * epsilonkl
    tmp2 = (1.0 - (zetax / 2.0)) / ((1.0 - zetax)**3)
    tmp3 = (9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax)**3))

    lx = len(Ikl)
    Bkl = np.zeros((len(rho), lx, lx))
    for i in np.arange(lx):
        for j in np.arange(lx):
            Bkl[:,i,j] = tmp11*tmp12[i,j]*(tmp2*Ikl[i,j] - tmp3*Jkl[i,j])

    return Bkl

def calc_a1ii(rho, Cmol2seg, dkl, l_akl, l_rkl, x0kl, epsilonkl, zetax):
    r"""
    Calculate effective first-order perturbation term :math:`\bar{a}_{1,ii}`.
    
    Used for the contribution of the monomeric interactions to the free energy per segment.
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)
    l_rkl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential repulsive exponents for k,l groups
    l_akl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential attractive exponents for k,l groups 
    x0kl : numpy.ndarray
        Matrix of sigmakl/dkl, ratio of Mie radius for groups scaled by hard sphere interaction (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    a1ii : numpy.ndarray
        Matrix used in the calculation of the radial distribution function of a hypothetical one-fluid Mie system.
    """
    Cii = (l_rkl / (l_rkl - l_akl)) * (l_rkl / l_akl)**(l_akl / (l_rkl - l_akl))

    Bii_r = calc_Bkl(rho, l_rkl, Cmol2seg, dkl, epsilonkl, x0kl, zetax)
    Bii_a = calc_Bkl(rho, l_akl, Cmol2seg, dkl, epsilonkl, x0kl, zetax)
    a1s_r = calc_a1s(rho, Cmol2seg, l_rkl, zetax, epsilonkl, dkl)
    a1s_a = calc_a1s(rho, Cmol2seg, l_akl, zetax, epsilonkl, dkl)

    return (Cii * (((x0kl**l_akl) / constants.molecule_per_nm3 * (a1s_a + Bii_a)) - ((x0kl**l_rkl) / constants.molecule_per_nm3 * (a1s_r + Bii_r))))

def calc_a1s_eff(rho, Cmol2seg, l_ii_avg, zetax, epsilonii_avg, dii_avg):
    r"""
    Return a1s,(rho*Cmol2seg,l_ij) in K
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.
    
    Returns
    -------
    a1s_eff : numpy.ndarray
        Term used in the used in the calculation of the effective averaged molecular radial distribution function of a hypothetical one-fluid Mie system.
    """
    
    ncomp = len(dii_avg)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1,4):
        zetax_pow[:, i] = zetax_pow[:, i-1] * zetax_pow[:, 0]
    
    # check if you have more than 1 bead types
    etaii_avg = np.zeros((len(rho), ncomp), dtype=rho.dtype)

    for k in range(ncomp):
        tmp = np.dot(ckl_coef, np.array( (1.0, 1.0/l_ii_avg[k], 1.0/l_ii_avg[k]**2, 1.0/l_ii_avg[k]**3), dtype=ckl_coef.dtype) )
        etaii_avg[:, k] = np.dot( zetax_pow, tmp )
    
    a1s = - (1.0 - (etaii_avg / 2.0)) / (1.0 - etaii_avg)**3 * 2.0 * np.pi * Cmol2seg * ((epsilonii_avg * (dii_avg**3 * constants.molecule_per_nm3**2)) / (l_ii_avg - 3.0) )
    
    output = np.transpose(np.transpose(a1s) * rho)
    
    return output

def calc_Bkl_eff(rho, l_ii_avg, Cmol2seg, dii_avg, epsilonii_avg, x0ii, zetax):
    r"""
    Return Bii_avg(rho*Cmol2seg,l_ii_avg) in K as defined in eq. 20.
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    Bii_avg : numpy.ndarray
        Bii_avg(rho*Cmol2seg,l_ii_avg) in K as defined in eq. 20.
    
    """
    
    rhos = Cmol2seg * rho
    ncomp = len(dii_avg)
    
    # compute Iii_avg(l_ii_avg), eq. 23
    Iii_avg = (1.0 - (x0ii**(3.0 - l_ii_avg))) / (l_ii_avg - 3.0)
    # compute Jii_avg(l_ii_avg), eq. 24
    Jii_avg = (1.0 - ((x0ii**(4.0 - l_ii_avg)) * (l_ii_avg - 3.0)) + ((x0ii**(3.0 - l_ii_avg)) * (l_ii_avg - 4.0))) / ((l_ii_avg - 3.0) * (l_ii_avg - 4.0))
    
    tmp11 = rhos * (2.0 * np.pi)
    tmp12 = (dii_avg**3 * constants.molecule_per_nm3**2) * epsilonii_avg
    tmp2 = (1.0 - (zetax / 2.0)) / ((1.0 - zetax)**3)
    tmp3 = (9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax)**3))
    
    Bii_avg = np.zeros((len(rho), ncomp))
    for k in np.arange(ncomp):
        Bii_avg[:,k] = tmp11*tmp12[k]*(tmp2*Iii_avg[k] - tmp3*Jii_avg[k])
    #  Bii_avg = tmp11*tmp12*(tmp2*Iii_avg - tmp3*Jii_avg)
    
    return Bii_avg

def calc_da1sii_drhos(rho, Cmol2seg, l_ii_avg, zetax, epsilonii_avg, dii_avg):
    r""" 
    Return a1s,ii_avg(rho*Cmol2seg,l_ii_avg) in K as defined in eq. 25.
    
    Used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.

    Returns
    -------
    calc_da1sii_drhos : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is the Ngroups by Ngroups
    """
    ncomp = len(dii_avg)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1,4):
        zetax_pow[:, i] = zetax_pow[:, i-1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etaii_avg = np.zeros((len(rho), ncomp), dtype=rho.dtype)
    rhos_detaii_avg_drhos = np.zeros((len(rho), ncomp), dtype=rho.dtype)

    for k in range(ncomp):
        ciii_avg = np.dot(ckl_coef, np.array( (1.0, 1.0/l_ii_avg[k], 1.0/l_ii_avg[k]**2, 1.0/l_ii_avg[k]**3), dtype=ckl_coef.dtype ) )
        etaii_avg[:, k] = np.dot( zetax_pow, ciii_avg)
        rhos_detaii_avg_drhos[:, k] = np.dot(zetax_pow, ciii_avg*np.array([1.,2.,3.,4.]))

    tmp1 = (1.0 - (etaii_avg / 2.0)) / ((1.0 - etaii_avg)**3) + (5.0-2.0*etaii_avg)/(2.0*(1.0-etaii_avg)**4)*rhos_detaii_avg_drhos
    tmp2 = - 2.0 * np.pi * ((epsilonii_avg * (dii_avg**3 * constants.molecule_per_nm3**2)) / (l_ii_avg - 3.0))

    da1s_drhos = tmp1*tmp2

    #da1s_drhos = - 2.0 * np.pi * ((1.0 - (etaii_avg / 2.0)) / ((1.0 - etaii_avg)**3) + (5.0 - 2.0*etaii_avg)/(2.0*(1.0-etaii_avg)**4)) * rhos_detaii_avg_drhos * ((epsilonii_avg * (dii_avg**3)) / (l_ii_avg - 3.0))
    return da1s_drhos

def calc_dBkl_drhos(l_ii_avg, dii_avg, epsilonii_avg, x0ii, zetax):
    r"""
    Return derivative of Bkl(rho*Cmol2seg,l_ii_avg) with respect to :math:`\rho_S`.
    
    Used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy.
    
    Parameters
    ----------
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e. molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    dBkl_drhos : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the perturbation expansion corresponding to the mean-attractive energy, size is rho x l_ii_avg.shape
    
    """
    ncomp = len(dii_avg)
    nrho = len(zetax)

    # compute Iii_avg(l_ii_avg), eq. 23
    Iii_avg = (1.0 - (x0ii**(3.0 - l_ii_avg))) / (l_ii_avg - 3.0)
    # compute Jii_avg(l_ii_avg), eq. 24
    Jii_avg = (1.0 - ((x0ii**(4.0 - l_ii_avg)) * (l_ii_avg - 3.0)) + ((x0ii**(3.0 - l_ii_avg)) * (l_ii_avg - 4.0))) / ((l_ii_avg - 3.0) * (l_ii_avg - 4.0))

    tmp = 2.0 * np.pi * dii_avg**3 * epsilonii_avg
    tmp1 = np.zeros((nrho,ncomp))
    tmp2 = np.zeros((nrho,ncomp))
    for k in np.arange(ncomp):
        tmp1[:,k] = ((1.0 - (zetax / 2.0)) / ((1.0 - zetax)**3) * Iii_avg[k]) - (((9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax)**3))) * Jii_avg[k])
        tmp2[:,k] = ((5.0 - 2.0*zetax) * zetax / (2*(1.0 - zetax)**4) * Iii_avg[k]) - (((9.0 * zetax * (zetax**2 + 4.0*zetax + 1)) / (2.0 * ((1 - zetax)**4))) * Jii_avg[k])
    dBkl_drhos = tmp*(tmp1 + tmp2) * constants.molecule_per_nm3**2

    return dBkl_drhos

def calc_da1iidrhos(rho, Cmol2seg, dii_eff, l_aii_avg, l_rii_avg, x0ii, epsilonii_avg, zetax):

    r"""
    Compute derivative of the term, :math:`\bar{a}_{1,ii}` with respect to :math:`\rho_s`
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dii_eff : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e. molecule) i.
    l_rii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e. molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    da1iidrhos : numpy.ndarray
        Derivative of term with respect to segment density
    """

    Cii = prefactor(l_rii_avg, l_aii_avg)

    das1_drhos_a = calc_da1sii_drhos(rho, Cmol2seg, l_aii_avg, zetax, epsilonii_avg, dii_eff)
    das1_drhos_r = calc_da1sii_drhos(rho, Cmol2seg, l_rii_avg, zetax, epsilonii_avg, dii_eff)

    dB_drhos_a = calc_dBkl_drhos(l_aii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dB_drhos_r = calc_dBkl_drhos(l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)

    da1iidrhos = (Cii * (((x0ii**l_aii_avg) * (das1_drhos_a + dB_drhos_a)) - ((x0ii**l_rii_avg) * (das1_drhos_r + dB_drhos_r))))

    return da1iidrhos

def calc_da2ii_1pchi_drhos(rho, Cmol2seg, epsilonii_avg, dii_eff, x0ii, l_rii_avg, l_aii_avg, zetax):

    r"""
    Compute derivative of the term, :math:`\frac{\bar{a}_{2,ii}}{1+\bar{\chi}_{ii}}` with respect to :math:`\rho_s`.
    
    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e. molecule) i.
    dii_eff : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in component (i.e. molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    l_rii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e. molecule) i.
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for groups (k,l)
    
    Returns
    -------
    da2ii_1pchi_drhos : numpy.ndarray
        Term used in the calculation of the second-order term from the macroscopic compressibility
    
    """

    # NoteHere g2mca der_a2kl

    # Calculate terms and derivatives used in derivative chain rule
    KHS = ((1.0 - zetax)**4) / (1.0 + (4.0 * zetax) + (4.0 * (zetax**2)) - (4.0 * (zetax**3)) + (zetax**4))
    dKHS_drhos = (4.0*(zetax**2 - 5.0*zetax - 2.0)*(1.0 - zetax)**3)/(zetax**4 - 4.0*zetax**3 + 4.0*zetax**2 + 4.0*zetax + 1.0)**2 *(zetax/(rho*Cmol2seg))

    a1sii_2l_aii_avg = calc_a1s_eff(rho, Cmol2seg, 2.0 * l_aii_avg, zetax, epsilonii_avg, dii_eff)
    a1sii_2l_rii_avg = calc_a1s_eff(rho, Cmol2seg, 2.0 * l_rii_avg, zetax, epsilonii_avg, dii_eff)
    a1sii_l_rii_avgl_aii_avg = calc_a1s_eff(rho, Cmol2seg, l_aii_avg + l_rii_avg, zetax, epsilonii_avg, dii_eff)

    Bii_2l_aii_avg = calc_Bkl_eff(rho, 2.0 * l_aii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)
    Bii_2l_rii_avg = calc_Bkl_eff(rho, 2.0 * l_rii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)
    Bii_l_aii_avgl_rii_avg = calc_Bkl_eff(rho, l_aii_avg + l_rii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)

    da1sii_2l_aii_avg = calc_da1sii_drhos(rho, Cmol2seg, 2.0 * l_aii_avg, zetax, epsilonii_avg, dii_eff)
    da1sii_2l_rii_avg = calc_da1sii_drhos(rho, Cmol2seg, 2.0 * l_rii_avg, zetax, epsilonii_avg, dii_eff)
    da1sii_l_rii_avgl_aii_avg = calc_da1sii_drhos(rho, Cmol2seg, l_aii_avg + l_rii_avg, zetax, epsilonii_avg, dii_eff)

    dBii_2l_aii_avg = calc_dBkl_drhos(2.0 * l_aii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dBii_2l_rii_avg = calc_dBkl_drhos(2.0 * l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dBii_l_aii_avgl_rii_avg = calc_dBkl_drhos(l_aii_avg + l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)

    # Calculate Derivative
    Cii = prefactor(l_rii_avg, l_aii_avg)

    B = x0ii**(2.0 * l_aii_avg) * (a1sii_2l_aii_avg + Bii_2l_aii_avg) - 2.0 * x0ii**(l_aii_avg + l_rii_avg) * (a1sii_l_rii_avgl_aii_avg + Bii_l_aii_avgl_rii_avg) + x0ii**(2.0 * l_rii_avg) * (a1sii_2l_rii_avg + Bii_2l_rii_avg)

    dA_B = np.transpose(np.transpose(0.5*epsilonii_avg*Cii**2 * B)*dKHS_drhos)
#    dA_B = np.einsum("i,ij->ij", dKHS_drhos, 0.5*epsilonii_avg*Cii**2 * B)

    dB = x0ii**(2.0 * l_aii_avg) * (da1sii_2l_aii_avg + dBii_2l_aii_avg) - 2.0 * x0ii**(l_aii_avg + l_rii_avg) * (da1sii_l_rii_avgl_aii_avg + dBii_l_aii_avgl_rii_avg) + x0ii**(2.0 * l_rii_avg) * (da1sii_2l_rii_avg + dBii_2l_rii_avg)

    A_dB = np.transpose(np.transpose(0.5*epsilonii_avg*Cii**2 * dB)*KHS)
    #A_dB = np.einsum("i,ij->ij", KHS, 0.5*epsilonii_avg*Cii**2 * dB)

    da2ii_1pchi_drhos = A_dB + dA_B

    return da2ii_1pchi_drhos




"""

The function names of this module represents global optimization methods that can be
specified as ``global_opts["method"]`` in :func:`~despasito.parameter_fitting.fit`.

"""

import os
import numpy as np
import logging
import scipy.optimize as spo

from despasito.utils.parallelization import MultiprocessingJob
from . import fit_functions as ff
import despasito.utils.general_toolbox as gtb

logger = logging.getLogger(__name__)


def single_objective(parameters_guess, bounds, fit_bead, fit_parameter_names, exp_dict, global_opts={}):
    r"""
    Evaluate parameter set for equation of state with given experimental data

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults from Eos object are broad,
        so we recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See
        EOS documentation for supported parameter names. Cross interaction parameter
        names should be composed of parameter name and the other bead type, separated
        by an underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional, default={}
        This dictionary is included for continuity with other global optimization
        methods, although this method doesn't have options.

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    if len(global_opts) > 0:
        logger.info("The fitting method 'single_objective' does not have further options")

    obj_value = ff.compute_obj(parameters_guess, fit_bead, fit_parameter_names, exp_dict, bounds)

    result = spo.OptimizeResult(
        x=parameters_guess,
        fun=obj_value,
        success=True,
        nit=0,
        message="Successfully computed objective function for provided parameter set.",
    )

    return result


def differential_evolution(
    parameters_guess,
    bounds,
    fit_bead,
    fit_parameter_names,
    exp_dict,
    global_opts={},
    constraints=None,
):
    r"""
    Fit defined parameters for equation of state object using
    scipy.optimize.differential_evolution with given experimental data.

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters. Not used in this method.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults from Eos object are broad,
        so we recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional

        - init (str) - Optional, default="random", type of initiation for population
        - write_intermediate_file (str) - Optional, default=False, If True, an
        intermediate file will be written from the method callback
        - filename (str) - Optional, default=None, filename for callback output, if
        provided, ``write_intermediate_file`` will be set to True
        - obj_cut (float) - Optional, default=None, Cut-off objective value to write
        the parameters, if provided, ``write_intermediate_file`` will be set to True
        - etc. Other keywords for scipy.optimize.differential_evolution use the
        function defaults

    constraints : dict, Optional, default=None
        This dictionary of constraint types and their arguments will be converted into
        a tuple of constraint classes that is compatible

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    global_opts = global_opts.copy()

    obj_kwargs = ["obj_cut", "filename", "write_intermediate_file"]
    if "obj_cut" in global_opts:
        obj_cut = global_opts["obj_cut"]
        del global_opts["obj_cut"]
        global_opts["write_intermediate_file"] = True
    else:
        obj_cut = None

    if "filename" in global_opts:
        filename = global_opts["filename"]
        del global_opts["filename"]
        global_opts["write_intermediate_file"] = True
    else:
        filename = None

    if "write_intermediate_file" in global_opts and global_opts["write_intermediate_file"]:
        del global_opts["write_intermediate_file"]
        global_opts["callback"] = _WriteParameterResults(fit_parameter_names, obj_cut=obj_cut, filename=filename)

    # Options for differential evolution, set defaults in new_global_opts
    new_global_opts = {"init": "random"}
    if global_opts:
        for key, value in global_opts.items():
            if key == "MultiprocessingObject":
                flag_workers = "workers" in global_opts and global_opts["workers"] > 1
                if value.ncores > 1 and flag_workers:
                    logger.info("Differential Evolution algorithm is using {} workers.".format(value.ncores))
                    new_global_opts["workers"] = value._pool.map
                    exp_dict = _del_Data_MultiprocessingObject(exp_dict)
            elif key not in obj_kwargs:
                new_global_opts[key] = value
    global_opts = new_global_opts

    if constraints is not None:
        global_opts["constraints"] = ff.initialize_constraints(constraints, "class")
    logger.info("Differential Evolution Options: {}".format(global_opts))

    result = spo.differential_evolution(
        ff.compute_obj, bounds, args=(fit_bead, fit_parameter_names, exp_dict, bounds), **global_opts
    )

    return result


def shgo(
    parameters_guess,
    bounds,
    fit_bead,
    fit_parameter_names,
    exp_dict,
    global_opts={},
    minimizer_opts={},
    constraints=None,
):
    r"""
    Fit defined parameters for equation of state object using scipy.optimize.shgo with
    given experimental data.

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults from Eos object are broad,
        so we recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional, default={}

        - init (str) - Optional, default="random", type of initiation for population
        - etc. Other keywords for scipy.optimize.shgo use the function defaults

    minimizer_opts : dict, Optional, default={}
        Dictionary used to define minimization type and the associated options.

        - method (str) - Optional, default=nelder-mead, Method available to
        scipy.optimize.minimize
        - options (dict) - Optional, default={'maxiter': 50}, This dictionary contains
        the kwargs available to the chosen method

    constraints : dict, Optional, default=None
        This dictionary of constraint types and their arguments will be converted into
        a tuple of dictionaries that is compatible

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    global_opts = global_opts.copy()

    # Options for differential evolution, set defaults in new_global_opts
    obj_kwargs = ["obj_cut", "filename", "write_intermediate_file"]
    new_global_opts = {"sampling_method": "sobol"}
    if global_opts:
        for key, value in global_opts.items():
            if key != "MultiprocessingObject" and key not in obj_kwargs:
                new_global_opts[key] = value
    global_opts = new_global_opts

    # Set up options for minimizer in basin hopping
    new_minimizer_opts = {"method": "nelder-mead", "options": {"maxiter": 50}}
    if minimizer_opts:
        for key, value in minimizer_opts.items():
            if key == "method":
                new_minimizer_opts[key] = value
            elif key == "options":
                for key2, value2 in value.items():
                    new_minimizer_opts[key][key2] = value2
    minimizer_opts = new_minimizer_opts

    if constraints is not None:
        global_opts["constraints"] = ff.initialize_constraints(constraints, "dict")
        if minimizer_opts["method"] not in ["COBYLA", "SLSQP"]:
            minimizer_opts["method"] = "SLSQP"
            for key, value in minimizer_opts["options"].items():
                if key not in [
                    "ftol",
                    "eps",
                    "disp",
                    "maxiter",
                    "finite_diff_rel_step",
                ]:
                    del minimizer_opts["options"][key]

    result = spo.shgo(
        ff.compute_obj,
        bounds,
        args=(fit_bead, fit_parameter_names, exp_dict, bounds),
        minimizer_kwargs=minimizer_opts,
        **global_opts,
    )

    return result


def grid_minimization(
    parameters_guess,
    bounds,
    fit_bead,
    fit_parameter_names,
    exp_dict,
    global_opts={},
    minimizer_opts={},
    constraints=None,
):
    r"""
    Fit defined parameters for equation of state object using a custom adaptation of
    scipy.optimize.brute with given experimental data.

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults from Eos object are broad,
        so we recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional, default={}

        - Ns (int) - Optional, default=5, Number of grid points along the axes
        - finish (callable) - Optional, default=scipy.optimize.minimize, A minimization
        function
        - initial_guesses (list) - Optional, Replaces grid of values generated with
        bounds and Ns
        - split_grid_minimization (int) - Optional, default=0, Choose index of first
        parameter to fit, while the grid is formed from those before. For example, if
        4 parameters are defined and ``split_grid_minimization==2``, then a grid is
        formed for the first two parameters ``parameters_guess[:2]``, and the remaining
        parameters, ``parameters_guess[2:]`` are minimized.

    minimizer_opts : dict, Optional, default={}
        Dictionary used to define minimization type and the associated options.

        - method (str) - Optional, default="least_squares", Method available to our
        :func:`~despasito.utils.general_toolbox.solve_root` function
        - options (dict) - Optional, default={}, This dictionary contains the kwargs
        available to the chosen method

    constraints : dict, Optional, default=None
        This dictionary of constraint types and their arguments will be converted into
        a tuple of constraint classes that is compatible

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    # Options for brute, set defaults in new_global_opts
    flag_use_mp_object = False
    new_global_opts = {"Ns": 5, "finish": spo.minimize, "split_grid_minimization": 0}
    if global_opts:
        for key, value in global_opts.items():
            if key == "MultiprocessingObject":
                if value.ncores > 1:
                    logger.info("Grid minimization algorithm is using {} workers.".format(value.ncores))
                    new_global_opts["MultiprocessingObject"] = value
                    flag_use_mp_object = True
                    exp_dict = _del_Data_MultiprocessingObject(exp_dict)
            else:
                new_global_opts[key] = value
    global_opts = new_global_opts

    if constraints is not None:
        global_opts["constraints"] = ff.initialize_constraints(constraints, "dict")

    logger.info("Grid Minimization Options: {}".format(global_opts))

    # Set up options for minimizer
    new_minimizer_opts = {"method": "least_squares"}
    if minimizer_opts:
        for key, value in minimizer_opts.items():
            if key == "method":
                new_minimizer_opts[key] = value
            elif key == "options":
                for key2, value2 in value.items():
                    new_minimizer_opts[key][key2] = value2
    minimizer_opts = new_minimizer_opts
    logger.info("    Minimizer Options: {}".format(minimizer_opts))

    args = (fit_bead, fit_parameter_names, exp_dict, bounds)

    # Set up inputs
    if "initial_guesses" in global_opts:
        del global_opts["Ns"]
        x0_array = global_opts["initial_guesses"]

        if global_opts["split_grid_minimization"] != 0:
            inputs = []
            bounds = bounds[global_opts["split_grid_minimization"]:]
            for x0 in x0_array:
                tmp1 = x0[global_opts["split_grid_minimization"]:]
                tmp2 = x0[: global_opts["split_grid_minimization"]]
                inputs.append((tmp1, (*args, tmp2), bounds, constraints, minimizer_opts))
        else:
            inputs = [(x0, args, bounds, constraints, minimizer_opts) for x0 in x0_array]

    else:
        # Initialization based on implementation in scipy.optimize.brute
        if global_opts["split_grid_minimization"] == 0:
            N = len(bounds)
            lrange = list(bounds)
            for k in range(N):
                if lrange[k] is not None:
                    if len(lrange[k]) < 3:
                        lrange[k] = tuple(lrange[k]) + (complex(global_opts["Ns"]),)
                    lrange[k] = slice(*lrange[k])
        else:
            if not isinstance(global_opts["split_grid_minimization"], int):
                raise ValueError("Option, split_grid_minimization, must be an integer")
            N = len(bounds[: global_opts["split_grid_minimization"]])
            lrange = list(bounds[: global_opts["split_grid_minimization"]])
            for k in range(N):
                if lrange[k] is not None:
                    if len(lrange[k]) < 3:
                        lrange[k] = tuple(lrange[k]) + (complex(global_opts["Ns"]),)
                    lrange[k] = slice(*lrange[k])
            bounds = bounds[global_opts["split_grid_minimization"]:]

        if N == 1:
            lrange = lrange[0]

        x0_array = np.mgrid[lrange]
        inpt_shape = x0_array.shape
        if N > 1:
            x0_array = np.reshape(x0_array, (inpt_shape[0], np.prod(inpt_shape[1:]))).T

        if global_opts["split_grid_minimization"] != 0:
            min_parameters = list(parameters_guess[global_opts["split_grid_minimization"]:])
            inputs = [(min_parameters, (*args, x0), bounds, constraints, minimizer_opts) for x0 in x0_array]
        else:
            inputs = [(x0, args, bounds, constraints, minimizer_opts) for x0 in x0_array]

    lx = len(x0_array)
    # Start computation
    if flag_use_mp_object:
        x0, results, fval = global_opts["MultiprocessingObject"].pool_job(_grid_minimization_wrapper, inputs)
    else:
        x0, results, fval = MultiprocessingJob.serial_job(_grid_minimization_wrapper, inputs)

    # Choose final output
    if global_opts["split_grid_minimization"] != 0:
        if "initial_guesses" not in global_opts:
            x0_new = np.zeros((lx, len(parameters_guess)))
        results_new = np.zeros((lx, len(parameters_guess)))
        for i in range(len(x0_array)):
            if "initial_guesses" not in global_opts:
                x0_new[i] = np.array(list(x0_array[i]) + list(min_parameters))
                results_new[i] = np.array(list(x0_array[i]) + list(results[i]))
            else:
                results_new[i] = np.array(
                    list(x0_array[i][: global_opts["split_grid_minimization"]]) + list(results[i])
                )
        results = results_new
        if "initial_guesses" not in global_opts:
            x0_array = x0_new

    result = [fval[0], results[0]]
    logger.info("For bead: {} and parameters {}".format(fit_bead, fit_parameter_names))
    for i in range(lx):
        tmp_result = results[i]
        logger.info("x0: {}, xf: {}, obj: {}".format(x0_array[i], tmp_result, fval[i]))
        if result[0] > fval[i]:
            result = [fval[i], tmp_result]

    result = spo.OptimizeResult(
        x=result[1],
        fun=result[0],
        success=True,
        nit=lx,
        message=(
            "Termination successful with {} grid points and".format(lx)
            + " the minimum value minimized. Note that parameters may be outside of"
            " the given bounds because of the minimizing function."
        ),
    )

    return result


def brute(parameters_guess, bounds, fit_bead, fit_parameter_names, exp_dict, global_opts={}):
    r"""
    Fit defined parameters for equation of state object using scipy.optimize.brute with
    given experimental data.

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum and
        maximum bounds of parameter being fit. Defaults from Eos object are broad, so we
        recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional, default={}

        - Ns (int) - Optional, default=5, Number of grid points along the axes
        - finish (callable) - Optional, default=scipy.optimize.minimize, An optimization
        function
        - etc. Other keywords for scipy.optimize.brute use the function defaults

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    # Options for brute, set defaults in new_global_opts
    new_global_opts = {"Ns": 5, "finish": spo.minimize}
    if global_opts:
        for key, value in global_opts.items():
            if key == "MultiprocessingObject":
                flag_workers = "workers" in global_opts and global_opts["workers"] > 1
                if value.ncores > 1 and flag_workers:
                    logger.info("Brute algorithm is using {} workers.".format(value.ncores))
                    new_global_opts["workers"] = value._pool.map
                    exp_dict = _del_Data_MultiprocessingObject(exp_dict)
            else:
                new_global_opts[key] = value
    global_opts = new_global_opts
    global_opts["full_output"] = True

    logger.info("Brute Options: {}".format(global_opts))
    x0, fval, grid, Jount = spo.brute(
        ff.compute_obj, bounds, args=(fit_bead, fit_parameter_names, exp_dict, bounds), **global_opts
    )
    result = spo.OptimizeResult(
        x=x0,
        fun=fval,
        success=True,
        nit=len(x0) * global_opts["Ns"],
        message=(
            "Termination successful with {} grid points and the minimum value "
            "minimized. Note that parameters may be outside of the given bounds "
            "because of the minimizing function.".format(len(x0) * global_opts["Ns"])
        ),
    )

    return result


def basinhopping(
    parameters_guess,
    bounds,
    fit_bead,
    fit_parameter_names,
    exp_dict,
    global_opts={},
    minimizer_opts={},
):
    r"""
    Fit defined parameters for equation of state object using
    scipy.optimize.basinhopping with given experimental data.

    Parameters
    ----------
    parameters_guess : numpy.ndarray
        An array of initial guesses for parameters.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum and
        maximum bounds of parameter being fit. Defaults from Eos object are broad, so
        we recommend specification.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional, default={}

        - niter (int) - Optional, default=10, The number of basin-hopping iterations
        - T (float) - Optional, default=0.5, The "temperature" parameter for the accept
        or reject criterion. For best results T should be comparable to the separation
        (in function value) between local minima.
        - niter_success (int) Optional, default=3, Stop the run if the global minimum
        candidate remains the same for this number of iterations.
        - stepsize (float) - Optional, default=0.1, Maximum step size for use in the
        random displacement.
        - take_step (callable) - Set with custom BasinStep class
        - write_intermediate_file (str) - Optional, default=False, If True, an
        intermediate file will be written from the method callback
        - filename (str) - Optional, default=None, filename for callback output, if
        provided, `write_intermediate_file` will be set to True
        - obj_cut (float) - Optional, default=None, Cut-off objective value to write
        the parameters, if provided, `write_intermediate_file` will be set to True
        - etc. Other keywords for scipy.optimize.basinhopping use the function
        defaults

    minimizer_opts : dict, Optional, default={}
        Dictionary used to define minimization type and the associated options.

        - method (str) - Method available to scipy.optimize.minimize
        - options (dict) - This dictionary contains the kwargs available to the chosen
        method

    Returns
    -------
    Objective : obj
        scipy OptimizedResult object

    """

    global_opts = global_opts.copy()

    if "obj_cut" in global_opts:
        obj_cut = global_opts["obj_cut"]
        del global_opts["obj_cut"]
        global_opts["write_intermediate_file"] = True
    else:
        obj_cut = None

    if "filename" in global_opts:
        filename = global_opts["filename"]
        del global_opts["filename"]
        global_opts["write_intermediate_file"] = True
    else:
        filename = None

    if "write_intermediate_file" in global_opts and global_opts["write_intermediate_file"]:
        del global_opts["write_intermediate_file"]
        global_opts["callback"] = _WriteParameterResults(fit_parameter_names, obj_cut=obj_cut, filename=filename)

    # Options for basin hopping
    new_global_opts = {"niter": 10, "T": 0.5, "niter_success": 3}
    if global_opts:
        for key, value in global_opts.items():
            if key != "MultiprocessingObject":
                new_global_opts[key] = value
    global_opts = new_global_opts

    # Set up options for minimizer in basin hopping
    new_minimizer_opts = {"method": "nelder-mead", "options": {"maxiter": 50}}
    if minimizer_opts:
        for key, value in minimizer_opts.items():
            if key == "method":
                new_minimizer_opts[key] = value
            elif key == "options":
                for key2, value2 in value.items():
                    new_minimizer_opts[key][key2] = value2
    minimizer_opts = new_minimizer_opts

    try:
        if "stepsize" in global_opts:
            stepsize = global_opts["stepsize"]
        else:
            stepsize = 0.1
        stepmag = np.transpose(np.array(bounds))[1]
        global_opts["take_step"] = _BasinStep(stepmag, stepsize=stepsize)
        custombounds = _BasinBounds(bounds)
    except Exception:
        raise TypeError("Could not initialize BasinStep and/or BasinBounds")

    logger.info("Basin Hopping Options: {}".format(global_opts))
    minimizer_kwargs = {
        "args": (fit_bead, fit_parameter_names, exp_dict, bounds),
        **minimizer_opts,
    }
    if "minimizer_kwargs" in global_opts:
        minimizer_kwargs.update(global_opts["minimizer_kwargs"])
        del global_opts["minimizer_kwargs"]
    result = spo.basinhopping(
        ff.compute_obj, parameters_guess, **global_opts, accept_test=custombounds, minimizer_kwargs=minimizer_kwargs
    )

    return result


# ___________ Supporting Classes and Functions _________________
def _grid_minimization_wrapper(args):
    """Wrapper for minimization method in grid_minimization"""

    x0, obj_args, bounds, constraints, opts = args

    if constraints is not None:
        logger.warning("Constraints defined, but grid_minimization does not support their use.")

    opts = opts.copy()
    if "method" in opts:
        method = opts["method"]
        del opts["method"]

    try:
        result = gtb.solve_root(
            ff.compute_obj,
            args=obj_args,
            method=method,
            x0=x0,
            bounds=bounds,
            options=opts,
        )
    except Exception:
        logger.info("Minimization Failed:", exc_info=True)
        result = np.nan * np.ones(len(x0)) if gtb.isiterable(x0) else np.nan

    logger.info("Starting parameters: {}, converged to: {}".format(x0, result))

    fval = ff.compute_obj(result, *obj_args)

    return x0, result, fval


class _BasinStep(object):
    r"""Custom basin step used by scipy.optimize.basinhopping function."""

    def __init__(self, stepmag, stepsize=0.05):
        r"""

        Parameters
        ----------
        stepmag : list
            List of step magnitudes
        stepsize : float, Optional, default=0.05
            Step size

        Attributes
        ----------
        stepmag : list
            List of step magnitudes
        stepsize : float, Optional, default=0.05
            Step size

        """

        self._stepsize = stepsize
        self._stepmag = stepmag

    def __call__(self, x):
        r"""

        Parameters
        ----------
        x : numpy.ndarray
            Guess in parameters values

        Returns
        -------
        basinstep : numpy.ndarray
            Suggested basin step used in scipy.optimize.basinhopping algorithm

        """

        # Save initial guess in array
        xold = np.copy(x)

        # Loop for number of times to start over
        for j in range(1000):
            # reset array x
            x = np.copy(xold)
            breakloop = True
            # Iterate through array of step magnitudes
            for i, mag in enumerate(self._stepmag):
                x[i] += np.random.uniform(-mag * self._stepsize, mag * self._stepsize)
                # If a value of x is negative, don't  break the cycle
                if x[i] < 0.0:
                    breakloop = False
            if breakloop:
                break
            logger.info("Basin Step after {} iterations:\n    {}".format(j, x))
        return x


class _BasinBounds(object):
    r"""Object used by scipy.optimize.basinhopping to set bounds of parameters."""

    def __init__(self, bounds):
        r"""

        Parameters
        ----------
        bounds : numpy.ndarray
            Bounds on parameters

        Attributes
        ----------
        xmin : numpy.ndarray
            Array of minimum values for each parameter
        xman : numpy.ndarray
            Array of maximum values for each parameter

        """
        bounds = np.transpose(np.array(bounds))
        self.xmin = bounds[0]
        self.xmax = bounds[1]

    def __call__(self, **kwargs):
        r"""

        Parameters
        ----------
        kwargs
            Keyword arguments used in BasinBounds object for
            scipy.optimize.basinhopping

            - x_new (numpy.ndarray) - Guess in parameters values
            - f_new (numpy.ndarray) - Objective value for given parameters

        Returns
        -------
        value : bool
            A true or false value that says whether the guess in parameter value is
            within bounds

        """
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))

        feasible1 = np.abs(kwargs["f_new"]) < np.inf
        feasible2 = not np.isnan(kwargs["f_new"])

        if tmax and tmin and feasible1 and feasible2:
            logger.info("Accept parameters: {}, with obj. function: {}".format(x, kwargs["f_new"]))
        else:
            logger.info("Reject parameters: {}, with obj. function: {}".format(x, kwargs["f_new"]))

        return tmax and tmin and feasible1 and feasible2


class _WriteParameterResults(object):
    r"""Object used by scipy.optimize.basinhopping to set bounds of parameters."""

    def __init__(self, beadnames, obj_cut=None, filename=None):
        r"""

        Attributes
        ----------
        beadnames : list[str]
            List of bead names for filename header
        obj_cut : float, Optional, default=np.inf
            Cut-off objective value to write the parameters
        filename : str, Optional, default=parameters.txt
            File to which parameters are written

        Returns
        -------
        Initiate file with parameters

        """

        if obj_cut is None:
            self.obj_cut = np.inf
        else:
            self.obj_cut = obj_cut

        if filename is None:
            filename = "parameters.txt"

        if os.path.isfile(filename):
            old_fname = filename
            for i in range(20):
                filename = "{}_{}".format(i, old_fname)
                if not os.path.isfile(filename):
                    logger.info("File '{}' already exists, using {}.".format(old_fname, filename))
                    break

        self.beadnames = beadnames
        self.filename = filename
        self.ninit = 0

    def __call__(self, *args, **kwargs):
        r"""
        The provided args and kwargs change depending on the global optimization
        method. This class is equipped to distinguish the callback function for
        differential_evolution (length equal to ) and basinhopping.

        Parameters
        ----------
        args
            The provided args change depending on the global optimization method.

            - x_new (numpy.ndarray) - Current parameter values being evaluated, used
            in both algorithms
            - f_new (float) - Current object function value for x_new, used in
            basinhopping
            - accept (bool) - Whether or not that minimum was accepted, used in
            basinhopping

        kwargs
            The provided kwargs change depending on the global optimization method.

            - convergence (float) - Used in differential evolution, the fractional
            value of the population convergence. When greater than one the function
            halts.

        Returns
        -------
        value : bool
            A true or false value that says whether the guess in parameter value is
            within bounds

        """

        if "convergence" in kwargs:  # Used in differential_evolution
            if kwargs["convergence"] < self.obj_cut:
                if not os.path.isfile(self.filename):
                    with open(self.filename, "w") as f:
                        f.write("# n, convergence, {}\n".format(", ".join(self.beadnames)))

                with open(self.filename, "a") as f:
                    tmp = [self.ninit, kwargs["convergence"]] + list(args[0])
                    f.write(("{}, " * len(tmp)).format(*tmp) + "\n")

        elif len(args) == 3:  # used in basinhopping
            if args[2] or args[1] < self.obj_cut:
                if not os.path.isfile(self.filename):
                    with open(self.filename, "w") as f:
                        f.write("# n, obj. value, accepted, {}\n".format(", ".join(self.beadnames)))

                with open(self.filename, "a") as f:
                    tmp = [self.ninit, args[1], args[2]] + list(args[0])
                    f.write(("{}, " * len(tmp)).format(*tmp) + "\n")
        else:
            raise ValueError(
                "Unknown inputs. This function is equipped to handle "
                "differential_evolution and basinhopping algorithms."
            )

        self.ninit += 1

        return False


def _del_Data_MultiprocessingObject(dictionary):
    r"""A dictionary of fitting objects will remove MultiprocessingObject attributes
    so that the multiprocessing pool can be used by the fitting algorithm.

    Parameters
    ----------
    dictionary : dict
        Dictionary of fitting objects

    Returns
    -------
    new_dictionary : dict
        Updated fitting objects
    """

    new_dictionary = dictionary.copy()
    for key in new_dictionary:
        if "MultiprocessingObject" in new_dictionary[key].thermodict:
            del new_dictionary[key].thermodict["MultiprocessingObject"]

    return new_dictionary



# -- coding: utf8 --
r"""
    EOS object for SAFT-:math:`\gamma`-Mie

    Equations referenced in this code are from V. Papaioannou et al J. Chem. Phys.
    140 054107 2014

"""

import sys
import numpy as np
import logging

import despasito.equations_of_state.eos_toolbox as tb
from despasito.equations_of_state import constants
import despasito.equations_of_state.saft.saft_toolbox as stb
from despasito.equations_of_state.saft import Aassoc
from .compiled_modules.ext_gamma_mie_python import prefactor, calc_Iij
from despasito.equations_of_state.saft.compiled_modules import ext_gamma_mie_numba as ext_numba
from despasito.equations_of_state.saft.compiled_modules import ext_gamma_mie_python as ext_python

logger = logging.getLogger(__name__)
if "cython" not in sys.modules:
    print("Cython package is unavailable, using Numba")
    flag_cython = True
else:
    flag_cython = False
    try:
        from despasito.equations_of_state.saft.compiled_modules import ext_gamma_mie_cython as ext_cython
    except ImportError:
        raise ImportError(
            "Cython package is available but module: "
            "despasito.equations_of_state.saft.compiled_modules.ext_Aassoc_cython, has"
            " not been compiled."
        )


class SaftType:
    r"""
    Object of SAFT-𝛾-Mie

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array containing the number of components by the number
        of bead types. Defines the number of each type of group in each component.
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - epsilon: :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann
        constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter [nm]
        - mass: Bead mass [kg/mol]
        - lambdar: :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups
        of type k
        - lambdaa: :math:`\lambda^{a}_{k,k}`, Exponent of attractive term between groups
        of type k
        - Sk: Optional, default=1, Shape factor, reflects the proportion with which a
        given segment contributes to the total free energy
        - Vks: Optional, default=1, Number of segments in this molecular group

    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of the
        desired parameters may be defined for whichever group combinations are desired.

        - epsilon: :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann
        constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter [nm]
        - mass: Bead mass [kg/mol]
        - lambdar: :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups
        of type k
        - lambdaa: :math:`\lambda^{a}_{k,k}`, Exponent of attractive term between groups
        of type k

    num_rings : list
        Number of rings in each molecule. This will impact the chain contribution to
        the Helmholtz energy.

    Attributes
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. See entry in **Parameters** section.
    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of
        the desired parameters may be defined for whichever group combinations are
        desired. Any interaction parameters that aren't provided are computed with the
        appropriate ``combining_rules``. See entry in **Parameters** section.
    Aideal_method : str
        "Abroglie" the default functional form of the ideal gas contribution of the
        Helmholtz energy
    residual_helmholtz_contributions : list[str]
        List of methods from the specified ``saft_source`` representing contributions
        to the Helmholtz energy that are functions of density, temperature, and
        composition. For this variant, [``Amonomer``, ``Achain``]
    parameter_types : list[str]
        This list of parameter names, "epsilon", "lambdar", "lambdaa", "sigma",
        and/or "Sk" as well as parameters for the main saft class.
    parameter_bound_extreme : dict
        With each parameter name as an entry representing a list with the minimum and
        maximum feasible parameter value.

        - epsilon: [100.,1000.]
        - lambdar: [6.0,100.]
        - lambdaa: [3.0,100.]
        - sigma: [0.1,10.0]
        - Sk: [0.1,1.0]

    combining_rules : dict
        Contains functional form and additional information for calculating cross
        interaction parameters that are not found in `cross_library`. Function must be
        one of those contained in
        :mod:`~despasito.equations_of_state.combining_rule_types`. The default values
        are:

        - sigma: {"function": "mean"}
        - lambdar: {"function": "mie_exponent"}
        - lambdar: {"function": "mie_exponent"}
        - epsilon: {"function": "volumetric_geometric_mean", "weighting_parameters":
        ["sigma"]}

    eos_dict : dict
        Dictionary of parameters and specific settings

        - molecular_composition (numpy.ndarray) - :math:`\nu_{i,k}/k_B`. Array
        containing the number of components by the number of bead types. Defines the
        number of each type of group in each component.
        - num_rings (list) - Number of rings in each molecule. This will impact the
        chain contribution to the Helmholtz energy.
        - Sk (numpy.ndarray) - Shape factor, reflects the proportion which a given
        segment contributes to the total free energy. Length of ``beads`` array.
        - Vks (numpy.ndarray) - Number of segments in this molecular group. Length of
        ``beads`` array.
        - Ckl (numpy.ndarray) - Matrix of Mie potential prefactors between beads (l,k)
        - epsilonkl (numpy.ndarray) - Matrix of Mie potential well depths for
        groups (k,l)
        - sigmakl (numpy.ndarray) - Matrix of bead diameters (k,l)
        - lambdarkl (numpy.ndarray) - Matrix of repulsive Mie exponent for
        groups (k,l)
        - lambdaakl (numpy.ndarray) - Matrix of attractive Mie exponent for
        groups (k,l)
        - dkl (numpy.ndarray) - Matrix of hard sphere equivalent for each bead and
        interaction between them (l,k)
        - x0kl (numpy.ndarray) - Matrix of sigmakl/dkl, sigmakl is the Mie radius for
        groups (k,l)
        - Cmol2seg (float) - Conversion factor from from molecular number density,
        :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.
        - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment
        or group) k multiplied by that of bead l
        - alphakl (np.array) - (Ngroup,Ngroup) "A dimensionless form of the integrated
        vdW energy of the Mie potential" eq. 33
        - epsilonii_avg (numpy.ndarray) - Matrix of molecule averaged well
        depths (i.j)
        - sigmaii_avg (numpy.ndarray) - Matrix of molecule averaged Mie
        diameter  (i.j)
        - lambdaaii_avg (numpy.ndarray) - Matrix of molecule averaged Mie potential
        attractive exponents  (i.j)
        - lambdarii_avg (numpy.ndarray) - Matrix of molecule averaged Mie potential
        attractive exponents (i.j)
        - dii_eff (numpy.ndarray) - Matrix of mole averaged hard sphere equivalent for
        each bead and interaction between them (i.j)
        - x0ii (numpy.ndarray) - Matrix of sigmaii_avg/dii_eff, sigmaii_avg is the
        average molecular Mie radius and dii_eff the average molecular hard sphere
        diameter

    ncomp : int
        Number of components in the system
    nbeads : int
        Number of beads in system that are shared among components
    xi : numpy.ndarray
        Mole fraction of each molecule in mixture. Default initialization is np.nan
    T : float
        Temperature value is initially defined as NaN for a placeholder until
        temperature dependent attributes are initialized by using a method of this
        class.

    """

    def __init__(self, **kwargs):

        if "method_stat" in kwargs:
            self.method_stat = kwargs["method_stat"]
            del kwargs["method_stat"]
        else:
            self.method_stat = None

        self.Aideal_method = "Abroglie"
        self.parameter_types = ["epsilon", "sigma", "lambdar", "lambdaa", "Sk"]
        self._parameter_defaults = {
            "epsilon": None,
            "lambdar": None,
            "lambdaa": None,
            "sigma": None,
            "Sk": 1.0,
            "Vks": 1.0,
        }
        self.parameter_bound_extreme = {
            "epsilon": [100.0, 1000.0],
            "sigma": [0.1, 1.0],
            "lambdar": [6.0, 100.0],
            "lambdaa": [3.0, 100.0],
            "Sk": [0.1, 1.0],
        }
        self.residual_helmholtz_contributions = ["Amonomer", "Achain"]
        self.combining_rules = {
            "sigma": {"function": "mean"},
            "lambdar": {"function": "mie_exponent"},
            "lambdaa": {"function": "mie_exponent"},
            "epsilon": {
                "function": "volumetric_geometric_mean",
                "weighting_parameters": ["sigma"],
            },
        }

        self._mixing_temp_dependence = None

        if not hasattr(self, "eos_dict"):
            self.eos_dict = {}

        needed_attributes = ["molecular_composition", "beads", "bead_library"]
        for key in needed_attributes:
            if key not in kwargs:
                raise ValueError("The one of the following inputs is missing: {}".format(", ".join(needed_attributes)))
            elif key == "molecular_composition":
                self.eos_dict[key] = kwargs[key]
            elif not hasattr(self, key):
                setattr(self, key, kwargs[key])

        self.bead_library = tb.check_bead_parameters(self.bead_library, self._parameter_defaults)

        if "cross_library" not in kwargs:
            self.cross_library = {}
        else:
            self.cross_library = kwargs["cross_library"]

        if "Vks" not in self.eos_dict:
            self.eos_dict["Vks"] = tb.extract_property("Vks", self.bead_library, self.beads, default=1.0)
        if "Sk" not in self.eos_dict:
            self.eos_dict["Sk"] = tb.extract_property("Sk", self.bead_library, self.beads, default=1.0)

        # Initialize temperature attribute
        if not hasattr(self, "T"):
            self.T = np.nan
        if not hasattr(self, "xi"):
            self.xi = np.nan

        if not hasattr(self, "nbeads") or not hasattr(self, "ncomp"):
            self.ncomp, self.nbeads = np.shape(self.eos_dict["molecular_composition"])

        # Initiate cross interaction terms
        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigmakl"] = output["sigma"]
        self.eos_dict["epsilonkl"] = output["epsilon"]
        self.eos_dict["lambdaakl"] = output["lambdaa"]
        self.eos_dict["lambdarkl"] = output["lambdar"]

        # compute alphakl eq. 33
        self.eos_dict["Ckl"] = prefactor(self.eos_dict["lambdarkl"], self.eos_dict["lambdaakl"])
        self.eos_dict["alphakl"] = self.eos_dict["Ckl"] * (
            (1.0 / (self.eos_dict["lambdaakl"] - 3.0)) - (1.0 / (self.eos_dict["lambdarkl"] - 3.0))
        )

        # Initiate average interaction terms
        self.calc_component_averaged_properties()

        if "num_rings" in kwargs:
            self.eos_dict["num_rings"] = kwargs["num_rings"]
            logger.info("Accepted component ring structure: {}".format(kwargs["num_rings"]))
        else:
            self.eos_dict["num_rings"] = np.zeros(len(self.eos_dict["molecular_composition"]))

    def calc_component_averaged_properties(self):
        r"""
        Calculate component averaged properties specific to SAFT-𝛾-Mie for the chain
        term.

        Attributes
        ----------
        output : dict
            Dictionary of outputs, the following possibilities are calculated if all
            relevant beads have those properties.

            - epsilonii_avg (numpy.ndarray) - Matrix of molecule averaged well depths
            - sigmaii_avg (numpy.ndarray) - Matrix of molecule averaged Mie diameter
            - lambdaaii_avg (numpy.ndarray) - Matrix of molecule averaged Mie potential
            attractive exponents
            - lambdarii_avg (numpy.ndarray) - Matrix of molecule averaged Mie potential
            attractive exponents

        """

        zki = np.zeros((self.ncomp, self.nbeads), float)
        zkinorm = np.zeros(self.ncomp, float)

        output = {}
        output["epsilonii_avg"] = np.zeros(self.ncomp, float)
        output["sigmaii_avg"] = np.zeros(self.ncomp, float)
        output["lambdarii_avg"] = np.zeros(self.ncomp, float)
        output["lambdaaii_avg"] = np.zeros(self.ncomp, float)

        # compute zki
        for i in range(self.ncomp):
            for k in range(self.nbeads):
                zki[i, k] = (
                    self.eos_dict["molecular_composition"][i, k] * self.eos_dict["Vks"][k] * self.eos_dict["Sk"][k]
                )
                zkinorm[i] += zki[i, k]

        for i in range(self.ncomp):
            for k in range(self.nbeads):
                zki[i, k] = zki[i, k] / zkinorm[i]

        for i in range(self.ncomp):
            for k in range(self.nbeads):
                for l in range(self.nbeads):
                    output["sigmaii_avg"][i] += zki[i, k] * zki[i, l] * self.eos_dict["sigmakl"][k, l] ** 3
                    output["epsilonii_avg"][i] += zki[i, k] * zki[i, l] * self.eos_dict["epsilonkl"][k, l]
                    output["lambdarii_avg"][i] += zki[i, k] * zki[i, l] * self.eos_dict["lambdarkl"][k, l]
                    output["lambdaaii_avg"][i] += zki[i, k] * zki[i, l] * self.eos_dict["lambdaakl"][k, l]
            output["sigmaii_avg"][i] = output["sigmaii_avg"][i] ** (1 / 3.0)

        self.eos_dict.update(output)

    def Ahard_sphere(self, rho, T, xi):
        r"""
        Outputs monomer contribution to the Helmholtz energy, :math:`A^{HS}/Nk_{B}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Ahard_sphere : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        eta = np.zeros((np.size(rho), 4))
        for m in range(4):
            eta[:, m] = (
                rho
                * constants.molecule_per_nm3
                * self.eos_dict["Cmol2seg"]
                * (
                    np.sum(np.sqrt(np.diag(self.eos_dict["xskl"])) * (np.diag(self.eos_dict["dkl"]) ** m))
                    * (np.pi / 6.0)
                )
            )

        tmp = 6.0 / (np.pi * rho * constants.molecule_per_nm3)
        if self.ncomp == 1:
            tmp1 = 0
        else:
            tmp1 = np.log1p(-eta[:, 3]) * (eta[:, 2] ** 3 / (eta[:, 3] ** 2) - eta[:, 0])
        tmp2 = 3.0 * eta[:, 2] / (1 - eta[:, 3]) * eta[:, 1]
        tmp3 = eta[:, 2] ** 3 / (eta[:, 3] * ((1.0 - eta[:, 3]) ** 2))

        AHS = tmp * (tmp1 + tmp2 + tmp3)

        return AHS

    def Afirst_order(self, rho, T, xi, zetax=None):
        r"""
        Outputs :math:`A^{1st order}/Nk_{B}T`. This is the first order term in the
        high-temperature perturbation expansion

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)

        Returns
        -------
        Afirst_order : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = stb.calc_zetax(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["dkl"],
            )

        # compute components of eq. 19
        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dkl"],
            self.eos_dict["lambdaakl"],
            self.eos_dict["lambdarkl"],
            self.eos_dict["x0kl"],
            self.eos_dict["epsilonkl"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            a1kl = ext_cython.calc_a1ii(*args)
        elif self.method_stat.python:
            a1kl = ext_python.calc_a1ii(*args)
        else:
            a1kl = ext_numba.calc_a1ii(*args)

        # eq. 18
        a1 = np.einsum("ijk,jk->i", a1kl, self.eos_dict["xskl"])
        A1 = (self.eos_dict["Cmol2seg"] / T) * a1  # Units of K

        return A1

    def Asecond_order(self, rho, T, xi, zetaxstar=None, zetax=None, KHS=None):
        r"""
        Outputs :math:`A^{2nd order}/Nk_{B}T`. This is the second order term in the
        high-temperature perturbation expansion

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetaxstar : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on sigma for groups (k,l)
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter for
            groups (k,l)
        KHS : numpy.ndarray, Optional, default=None
            (length of densities) isothermal compressibility of system with packing
            fraction zetax

        Returns
        -------
        Asecond_order : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = stb.calc_zetax(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["dkl"],
            )

        if zetaxstar is None:
            zetaxstar = stb.calc_zetaxstar(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["sigmakl"],
            )

        if KHS is None:
            KHS = stb.calc_KHS(zetax)

        # compute a2kl, eq. 30 #####

        # compute f1, f2, and f3 for eq. 32
        fmlist123 = self.calc_fm(self.eos_dict["alphakl"], np.array([1, 2, 3]))

        chikl = (
            np.einsum("i,jk", zetaxstar, fmlist123[0])
            + np.einsum("i,jk", zetaxstar**5, fmlist123[1])
            + np.einsum("i,jk", zetaxstar**8, fmlist123[2])
        )

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            2.0 * self.eos_dict["lambdaakl"],
            zetax,
            self.eos_dict["epsilonkl"],
            self.eos_dict["dkl"],
        )
        if self.method_stat.cython and flag_cython:
            a1s_2la = ext_cython.calc_a1s(*args)
        elif self.method_stat.python:
            a1s_2la = ext_python.calc_a1s(*args)
        else:
            a1s_2la = ext_numba.calc_a1s(*args)
        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            2.0 * self.eos_dict["lambdarkl"],
            zetax,
            self.eos_dict["epsilonkl"],
            self.eos_dict["dkl"],
        )
        if self.method_stat.cython and flag_cython:
            a1s_2lr = ext_cython.calc_a1s(*args)
        elif self.method_stat.python:
            a1s_2lr = ext_python.calc_a1s(*args)
        else:
            a1s_2lr = ext_numba.calc_a1s(*args)
        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["lambdaakl"] + self.eos_dict["lambdarkl"],
            zetax,
            self.eos_dict["epsilonkl"],
            self.eos_dict["dkl"],
        )
        if self.method_stat.cython and flag_cython:
            a1s_lalr = ext_cython.calc_a1s(*args)
        elif self.method_stat.python:
            a1s_lalr = ext_python.calc_a1s(*args)
        else:
            a1s_lalr = ext_numba.calc_a1s(*args)
        args = (
            rho,
            2.0 * self.eos_dict["lambdaakl"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dkl"],
            self.eos_dict["epsilonkl"],
            self.eos_dict["x0kl"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            B_2la = ext_cython.calc_Bkl(*args)
        elif self.method_stat.python:
            B_2la = ext_python.calc_Bkl(*args)
        else:
            B_2la = ext_numba.calc_Bkl(*args)
        args = (
            rho,
            2.0 * self.eos_dict["lambdarkl"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dkl"],
            self.eos_dict["epsilonkl"],
            self.eos_dict["x0kl"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            B_2lr = ext_cython.calc_Bkl(*args)
        elif self.method_stat.python:
            B_2lr = ext_python.calc_Bkl(*args)
        else:
            B_2lr = ext_numba.calc_Bkl(*args)
        args = (
            rho,
            self.eos_dict["lambdaakl"] + self.eos_dict["lambdarkl"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dkl"],
            self.eos_dict["epsilonkl"],
            self.eos_dict["x0kl"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            B_lalr = ext_cython.calc_Bkl(*args)
        elif self.method_stat.python:
            B_lalr = ext_python.calc_Bkl(*args)
        else:
            B_lalr = ext_numba.calc_Bkl(*args)

        a2kl = (
            (self.eos_dict["x0kl"] ** (2.0 * self.eos_dict["lambdaakl"]))
            * (a1s_2la + B_2la)
            / constants.molecule_per_nm3
            - (
                (2.0 * self.eos_dict["x0kl"] ** (self.eos_dict["lambdaakl"] + self.eos_dict["lambdarkl"]))
                * (a1s_lalr + B_lalr)
                / constants.molecule_per_nm3
            )
            + (
                (self.eos_dict["x0kl"] ** (2.0 * self.eos_dict["lambdarkl"]))
                * (a1s_2lr + B_2lr)
                / constants.molecule_per_nm3
            )
        )
        a2kl *= (1.0 + chikl) * self.eos_dict["epsilonkl"] * (self.eos_dict["Ckl"] ** 2)  # *(KHS/2.0)

        a2kl = np.einsum("i,ijk->ijk", KHS / 2.0, a2kl)

        # eq. 29
        a2 = np.einsum("ijk,jk->i", a2kl, self.eos_dict["xskl"])
        A2 = (self.eos_dict["Cmol2seg"] / (T**2)) * a2

        return A2

    def Athird_order(self, rho, T, xi, zetaxstar=None):
        r"""
        Outputs :math:`A^{3rd order}/Nk_{B}T`. This is the third order term in the
        high-temperature perturbation expansion

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetaxstar : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on sigma for groups (k,l)

        Returns
        -------
        Athird_order : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetaxstar is None:
            zetaxstar = stb.calc_zetaxstar(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["sigmakl"],
            )

        # compute a3kl
        fmlist456 = self.calc_fm(self.eos_dict["alphakl"], np.array([4, 5, 6]))

        a3kl = np.einsum("i,jk", zetaxstar, -(self.eos_dict["epsilonkl"] ** 3) * fmlist456[0]) * np.exp(
            np.einsum("i,jk", zetaxstar, fmlist456[1]) + np.einsum("i,jk", zetaxstar**2, fmlist456[2])
        )

        # eq. 37
        a3 = np.einsum("ijk,jk->i", a3kl, self.eos_dict["xskl"])
        A3 = (self.eos_dict["Cmol2seg"] / (T**3)) * a3

        return A3

    def Amonomer(self, rho, T, xi):
        r"""
        Outputs the monomer contribution of the Helmholtz energy,
        :math:`A^{mono.}/Nk_{B}T`.

        This term is composed of:
        :math:`A^{HS}/Nk_{B}T + A^{1st order}/Nk_{B}T + A^{2nd order}/Nk_{B}T` +
        :math:`A^{3rd order}/Nk_{B}T`

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Amonomer : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        if np.all(rho >= self.density_max(xi, T, maxpack=1.0)):
            raise ValueError(
                "Density values should not all be greater than "
                + "{},".format(self.density_max(xi, T))
                + " or calc_Amono will fail in log calculation."
            )

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        zetax = stb.calc_zetax(rho, self.eos_dict["Cmol2seg"], self.eos_dict["xskl"], self.eos_dict["dkl"])
        zetaxstar = stb.calc_zetaxstar(
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["xskl"],
            self.eos_dict["sigmakl"],
        )
        Amonomer = (
            self.Ahard_sphere(rho, T, xi)
            + self.Afirst_order(rho, T, xi, zetax=zetax)
            + self.Asecond_order(rho, T, xi, zetax=zetax, zetaxstar=zetaxstar)
            + self.Athird_order(rho, T, xi, zetaxstar=zetaxstar)
        )

        return Amonomer

    def gdHS(self, rho, T, xi, zetax=None):
        r"""
        The zeroth order expansion term in calculating the radial distribution
        function of a Mie fluid.

        This is also known as the hard sphere radial distribution function.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter for
            groups (k,l)

        Returns
        -------
        gdHS : numpy.ndarray
            Hard sphere radial distribution function
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = stb.calc_zetax(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["dkl"],
            )

        km = np.zeros((np.size(rho), 4))
        gdHS = np.zeros((np.size(rho), np.size(xi)))

        km[:, 0] = -np.log(1.0 - zetax) + (42.0 * zetax - 39.0 * zetax**2 + 9.0 * zetax**3 - 2.0 * zetax**4) / (
            6.0 * (1.0 - zetax) ** 3
        )
        km[:, 1] = (zetax**4 + 6.0 * zetax**2 - 12.0 * zetax) / (2.0 * (1.0 - zetax) ** 3)
        km[:, 2] = -3.0 * zetax**2 / (8.0 * (1.0 - zetax) ** 2)
        km[:, 3] = (-(zetax**4) + 3.0 * zetax**2 + 3.0 * zetax) / (6.0 * (1.0 - zetax) ** 3)

        for i in range(self.ncomp):
            gdHS[:, i] = np.exp(
                km[:, 0]
                + km[:, 1] * self.eos_dict["x0ii"][i]
                + km[:, 2] * self.eos_dict["x0ii"][i] ** 2
                + km[:, 3] * self.eos_dict["x0ii"][i] ** 3
            )

        return gdHS

    def g1(self, rho, T, xi, zetax=None):
        r"""
        Calculate the first order expansion term in calculating the radial
        distribution function of a Mie fluid

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)

        Returns
        -------
        g1 : numpy.ndarray
            First order expansion term in calculating the radial distribution
            function of a Mie fluid
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = stb.calc_zetax(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["dkl"],
            )

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["lambdaaii_avg"],
            self.eos_dict["lambdarii_avg"],
            self.eos_dict["x0ii"],
            self.eos_dict["epsilonii_avg"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            da1iidrhos = ext_cython.calc_da1iidrhos(*args)
        elif self.method_stat.python:
            da1iidrhos = ext_python.calc_da1iidrhos(*args)
        else:
            da1iidrhos = ext_numba.calc_da1iidrhos(*args)

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["lambdaaii_avg"],
            zetax,
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
        )
        if self.method_stat.cython and flag_cython:
            a1sii_lambdaaii_avg = ext_cython.calc_a1s_eff(*args)
        elif self.method_stat.python:
            a1sii_lambdaaii_avg = ext_python.calc_a1s_eff(*args)
        else:
            a1sii_lambdaaii_avg = ext_numba.calc_a1s_eff(*args)

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["lambdarii_avg"],
            zetax,
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
        )
        if self.method_stat.cython and flag_cython:
            a1sii_lambdarii_avg = ext_cython.calc_a1s_eff(*args)
        elif self.method_stat.python:
            a1sii_lambdarii_avg = ext_python.calc_a1s_eff(*args)
        else:
            a1sii_lambdarii_avg = ext_numba.calc_a1s_eff(*args)

        args = (
            rho,
            self.eos_dict["lambdaaii_avg"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["x0ii"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            Bii_lambdaaii_avg = ext_cython.calc_Bkl_eff(*args)
        elif self.method_stat.python:
            Bii_lambdaaii_avg = ext_python.calc_Bkl_eff(*args)
        else:
            Bii_lambdaaii_avg = ext_numba.calc_Bkl_eff(*args)

        args = (
            rho,
            self.eos_dict["lambdarii_avg"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["x0ii"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            Bii_lambdarii_avg = ext_cython.calc_Bkl_eff(*args)
        elif self.method_stat.python:
            Bii_lambdarii_avg = ext_python.calc_Bkl_eff(*args)
        else:
            Bii_lambdarii_avg = ext_numba.calc_Bkl_eff(*args)

        Cii = prefactor(self.eos_dict["lambdarii_avg"], self.eos_dict["lambdaaii_avg"])

        tmp1 = 1.0 / (
            2.0 * np.pi * self.eos_dict["epsilonii_avg"] * self.eos_dict["dii_eff"] ** 3 * constants.molecule_per_nm3**2
        )
        tmp11 = 3.0 * da1iidrhos
        tmp21 = Cii * self.eos_dict["lambdaaii_avg"] * (self.eos_dict["x0ii"] ** self.eos_dict["lambdaaii_avg"])
        tmp22 = np.einsum(
            "ij,i->ij",
            (a1sii_lambdaaii_avg + Bii_lambdaaii_avg),
            1.0 / (rho * self.eos_dict["Cmol2seg"]),
        )
        tmp31 = Cii * self.eos_dict["lambdarii_avg"] * (self.eos_dict["x0ii"] ** self.eos_dict["lambdarii_avg"])
        tmp32 = np.einsum(
            "ij,i->ij",
            (a1sii_lambdarii_avg + Bii_lambdarii_avg),
            1.0 / (rho * self.eos_dict["Cmol2seg"]),
        )
        g1 = tmp1 * (tmp11 - tmp21 * tmp22 + tmp31 * tmp32)

        return g1

    def g2(self, rho, T, xi, zetax=None):
        r"""
        Calculate the second order expansion term in calculating the radial
        distribution function of a Mie fluid

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)

        Returns
        -------
        g2 : numpy.ndarray
            Second order expansion term in calculating the radial distribution
            function of a Mie fluid
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = stb.calc_zetax(
                rho,
                self.eos_dict["Cmol2seg"],
                self.eos_dict["xskl"],
                self.eos_dict["dkl"],
            )
        zetaxstar = stb.calc_zetaxstar(
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["xskl"],
            self.eos_dict["sigmakl"],
        )
        KHS = stb.calc_KHS(zetax)

        Cii = prefactor(self.eos_dict["lambdarii_avg"], self.eos_dict["lambdaaii_avg"])

        phi7 = np.array([10.0, 10.0, 0.57, -6.7, -8.0])
        alphaii = Cii * (
            (1.0 / (self.eos_dict["lambdaaii_avg"] - 3.0)) - (1.0 / (self.eos_dict["lambdarii_avg"] - 3.0))
        )
        theta = np.exp(self.eos_dict["epsilonii_avg"] / T) - 1.0

        gammacii = np.zeros((np.size(rho), np.size(xi)))
        for i in range(self.ncomp):
            gammacii[:, i] = (
                phi7[0]
                * (-np.tanh(phi7[1] * (phi7[2] - alphaii[i])) + 1.0)
                * zetaxstar
                * theta[i]
                * np.exp(phi7[3] * zetaxstar + phi7[4] * (zetaxstar**2))
            )

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["x0ii"],
            self.eos_dict["lambdarii_avg"],
            self.eos_dict["lambdaaii_avg"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            da2iidrhos = ext_cython.calc_da2ii_1pchi_drhos(*args)
        elif self.method_stat.python:
            da2iidrhos = ext_python.calc_da2ii_1pchi_drhos(*args)
        else:
            da2iidrhos = ext_numba.calc_da2ii_1pchi_drhos(*args)

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            2.0 * self.eos_dict["lambdaaii_avg"],
            zetax,
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
        )
        if self.method_stat.cython and flag_cython:
            a1sii_2lambdaaii_avg = ext_cython.calc_a1s_eff(*args)
        elif self.method_stat.python:
            a1sii_2lambdaaii_avg = ext_python.calc_a1s_eff(*args)
        else:
            a1sii_2lambdaaii_avg = ext_numba.calc_a1s_eff(*args)

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            2.0 * self.eos_dict["lambdarii_avg"],
            zetax,
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
        )
        if self.method_stat.cython and flag_cython:
            a1sii_2lambdarii_avg = ext_cython.calc_a1s_eff(*args)
        elif self.method_stat.python:
            a1sii_2lambdarii_avg = ext_python.calc_a1s_eff(*args)
        else:
            a1sii_2lambdarii_avg = ext_numba.calc_a1s_eff(*args)

        args = (
            rho,
            self.eos_dict["Cmol2seg"],
            self.eos_dict["lambdaaii_avg"] + self.eos_dict["lambdarii_avg"],
            zetax,
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["dii_eff"],
        )
        if self.method_stat.cython and flag_cython:
            a1sii_lambdarii_avglambdaaii_avg = ext_cython.calc_a1s_eff(*args)
        elif self.method_stat.python:
            a1sii_lambdarii_avglambdaaii_avg = ext_python.calc_a1s_eff(*args)
        else:
            a1sii_lambdarii_avglambdaaii_avg = ext_numba.calc_a1s_eff(*args)

        args = (
            rho,
            2.0 * self.eos_dict["lambdaaii_avg"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["x0ii"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            Bii_2lambdaaii_avg = ext_cython.calc_Bkl_eff(*args)
        elif self.method_stat.python:
            Bii_2lambdaaii_avg = ext_python.calc_Bkl_eff(*args)
        else:
            Bii_2lambdaaii_avg = ext_numba.calc_Bkl_eff(*args)

        args = (
            rho,
            2.0 * self.eos_dict["lambdarii_avg"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["x0ii"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            Bii_2lambdarii_avg = ext_cython.calc_Bkl_eff(*args)
        elif self.method_stat.python:
            Bii_2lambdarii_avg = ext_python.calc_Bkl_eff(*args)
        else:
            Bii_2lambdarii_avg = ext_numba.calc_Bkl_eff(*args)

        args = (
            rho,
            self.eos_dict["lambdaaii_avg"] + self.eos_dict["lambdarii_avg"],
            self.eos_dict["Cmol2seg"],
            self.eos_dict["dii_eff"],
            self.eos_dict["epsilonii_avg"],
            self.eos_dict["x0ii"],
            zetax,
        )
        if self.method_stat.cython and flag_cython:
            Bii_lambdaaii_avglambdarii_avg = ext_cython.calc_Bkl_eff(*args)
        elif self.method_stat.python:
            Bii_lambdaaii_avglambdarii_avg = ext_python.calc_Bkl_eff(*args)
        else:
            Bii_lambdaaii_avglambdarii_avg = ext_numba.calc_Bkl_eff(*args)

        eKC2 = np.einsum(
            "i,j->ij",
            KHS / rho / self.eos_dict["Cmol2seg"],
            self.eos_dict["epsilonii_avg"] * (Cii**2),
        )

        g2MCA = (
            1.0
            / (
                2.0
                * np.pi
                * (self.eos_dict["epsilonii_avg"] ** 2)
                * self.eos_dict["dii_eff"] ** 3
                * constants.molecule_per_nm3**2
            )
        ) * (
            (3.0 * da2iidrhos)
            - (
                eKC2
                * self.eos_dict["lambdarii_avg"]
                * (self.eos_dict["x0ii"] ** (2.0 * self.eos_dict["lambdarii_avg"]))
            )
            * (a1sii_2lambdarii_avg + Bii_2lambdarii_avg)
            + eKC2
            * (self.eos_dict["lambdarii_avg"] + self.eos_dict["lambdaaii_avg"])
            * (self.eos_dict["x0ii"] ** (self.eos_dict["lambdarii_avg"] + self.eos_dict["lambdaaii_avg"]))
            * (a1sii_lambdarii_avglambdaaii_avg + Bii_lambdaaii_avglambdarii_avg)
            - eKC2
            * self.eos_dict["lambdaaii_avg"]
            * (self.eos_dict["x0ii"] ** (2.0 * self.eos_dict["lambdaaii_avg"]))
            * (a1sii_2lambdaaii_avg + Bii_2lambdaaii_avg)
        )

        g2 = (1.0 + gammacii) * g2MCA

        return g2

    def Achain(self, rho, T, xi):
        r"""
        Outputs the chain term for the Helmholtz energy, :math:`A^{chain}/Nk_{B}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Achain : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        zetax = stb.calc_zetax(rho, self.eos_dict["Cmol2seg"], self.eos_dict["xskl"], self.eos_dict["dkl"])
        gdHS = self.gdHS(rho, T, xi, zetax=zetax)
        g1 = self.g1(rho, T, xi, zetax=zetax)
        g2 = self.g2(rho, T, xi, zetax=zetax)

        gii = gdHS * np.exp(
            (self.eos_dict["epsilonii_avg"] * g1 / (T * gdHS))
            + (((self.eos_dict["epsilonii_avg"] / T) ** 2) * g2 / gdHS)
        )

        Achain = 0.0
        for i in range(self.ncomp):
            beadsum = -1.0 + self.eos_dict["num_rings"][i]
            for k in range(self.nbeads):
                beadsum += (
                    self.eos_dict["molecular_composition"][i, k] * self.eos_dict["Vks"][k] * self.eos_dict["Sk"][k]
                )
            Achain -= xi[i] * beadsum * np.log(gii[:, i])

        if np.any(np.isnan(Achain)):
            logger.error("Some Helmholtz values are NaN, check energy parameters.")

        return Achain

    def density_max(self, xi, T, maxpack=0.65):
        """
        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.65
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [:math:`mol/m^3`]
        """

        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        # estimate the maximum density based on the hard sphere packing fraction
        # etax, assuming a maximum packing fraction specified by maxpack

        max_density = (
            maxpack
            * 6.0
            / (self.eos_dict["Cmol2seg"] * np.pi * np.sum(self.eos_dict["xskl"] * (self.eos_dict["dkl"] ** 3)))
            / constants.molecule_per_nm3
        )

        return max_density

    @staticmethod
    def calc_fm(alphakl, mlist):
        r"""
        Calculate list of coefficients used to compute the correction term for
        :math:`A_{2nd order}/Nk_{B}T` which is related to the fluctuations of
        attractive energy. where a list of m values are specified in mlist eq. 39

        Parameters
        ----------
        alphakl : numpy.ndarray
            (Ngroup,Ngroup) "A dimensionless form of the integrated vdW energy of
            the Mie potential" eq. 33
        mlist : numpy.ndarray
            (number of m values) an array of integers used in the calculation of
            :math:`A^{mono}`

        Returns
        -------
        fmlist : numpy.ndarray
            List of coefficients used to compute the correction term for
            :math:`A_{2}` which is related to the fluctuations of attractive energy.
        """

        if np.size(np.shape(alphakl)) == 2:
            fmlist = np.zeros((np.size(mlist), np.size(alphakl, axis=0), np.size(alphakl, axis=0)))
        elif np.size(np.shape(alphakl)) == 1:
            fmlist = np.zeros((np.size(mlist), np.size(alphakl, axis=0)))
        else:
            logger.error("Unexpected shape in calc_fm")
        mlist = mlist - 1

        phimn = np.array(
            [
                [
                    7.53655570e00,
                    -3.76046300e01,
                    7.17459530e01,
                    -4.68355200e01,
                    -2.46798200e00,
                    -5.02720000e-01,
                    8.09568830e00,
                ],
                [
                    -3.59440000e02,
                    1.82560000e03,
                    -3.16800000e03,
                    1.88420000e03,
                    -8.23760000e-01,
                    -3.19350000e00,
                    3.70900000e00,
                ],
                [
                    1.55090000e03,
                    -5.07010000e03,
                    6.53460000e03,
                    -3.28870000e03,
                    -2.71710000e00,
                    2.08830000e00,
                    0.00000000e00,
                ],
                [
                    -1.19932000e00,
                    9.06363200e00,
                    -1.79482000e01,
                    1.13402700e01,
                    2.05214200e01,
                    -5.66377000e01,
                    4.05368300e01,
                ],
                [
                    -1.91128000e03,
                    2.13901750e04,
                    -5.13207000e04,
                    3.70645400e04,
                    1.10374200e03,
                    -3.26461000e03,
                    2.55618100e03,
                ],
                [
                    9.23690000e03,
                    -1.29430000e05,
                    3.57230000e05,
                    -3.15530000e05,
                    1.39020000e03,
                    -4.51820000e03,
                    4.24160000e03,
                ],
                [
                    1.00000000e01,
                    1.00000000e01,
                    5.70000000e-01,
                    -6.70000000e00,
                    -8.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                ],
            ]
        )

        for i, m in enumerate(mlist):
            for n in range(4):
                fmlist[i] += phimn[m, n] * (alphakl**n)
            dum = np.ones_like(fmlist[i])
            for n in range(4, 7):
                dum += phimn[m, n] * (alphakl ** (n - 3.0))
            fmlist[i] = fmlist[i] / dum

        return fmlist

    def calc_gr_assoc(self, rho, T, xi, Ktype="ijklab"):
        r"""

        Reference fluid pair correlation function used in calculating association
        sites

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        Ktype : str, Optional, default='ijklab'
            Indicates which radial distribution function to return

            - 'ijklab': The bonding volume was calculated from self.calc_Kijklab,
            return gHS_dij)
            - 'klab': The bonding volume was provided to saft.py so use
            temperature-density polynomial correlation

        Returns
        -------
        Iij : numpy.ndarray
            A temperature-density polynomial correlation of the association integral
            for a Lennard-Jones monomer. This matrix is (len(rho) x Ncomp x Ncomp)
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        if Ktype == "klab":
            gr = calc_Iij(
                rho,
                T,
                xi,
                self.eos_dict["epsilonii_avg"],
                self.eos_dict["sigmaii_avg"],
                self.eos_dict["sigmakl"],
                self.eos_dict["xskl"],
            )
        elif Ktype == "ijklab":
            gr = self.calc_gdHS_assoc(rho, T, xi)
        else:
            raise ValueError("Ktype does not indicate a known gr_assoc for this saft type.")

        return gr

    def calc_gdHS_assoc(self, rho, T, xi):
        r"""

        Radial distribution function at contact.

        Papaioannou J. Chem. Phys. 140, 054107 (2014)

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        gr : numpy.ndarray
            This matrix is (len(rho) x Ncomp x Ncomp)
        """

        rho = self._check_density(rho)
        self._check_temperature_dependent_parameters(T)
        self._check_composition_dependent_parameters(xi)

        eta = np.zeros((np.size(rho), 2))
        for m in range(2, 4):
            eta[:, m] = (
                rho
                * constants.molecule_per_nm3
                * self.eos_dict["Cmol2seg"]
                * (
                    np.sum(np.sqrt(np.diag(self.eos_dict["xskl"])) * (np.diag(self.eos_dict["dkl"]) ** m))
                    * (np.pi / 6.0)
                )
            )

        gr = np.zeros((len(rho), self.ncomp, self.ncomp))
        tmp0 = 1 / (1 - eta[:, 1])
        tmp1 = eta[:, 0] / (1 - eta[:, 1]) ** 2
        tmp2 = eta[:, 0] ** 2 / (1 - eta[:, 1]) ** 3
        for i in range(self.ncomp):
            for j in range(self.ncomp):
                tmp = (
                    self.eos_dict["dii_eff"][i]
                    * self.eos_dict["dii_eff"][j]
                    / (self.eos_dict["dii_eff"][i] + self.eos_dict["dii_eff"][j])
                )
                gr[:, i, j] = tmp0 + 3 * tmp * tmp1 + 2 * tmp**2 * tmp2

        return gr

    def calc_Kijklab(self, T, rc_klab, rd_klab=None, reduction_ratio=0.25):
        r"""

        Calculation of association site bonding volume, dependent on molecule in
        addition to group

        Papaioannou J. Chem. Phys. 140, 054107 (2014)

        Parameters
        ----------
        T : float
            Temperature of the system [K]
        rc_klab : numpy.ndarray
            This matrix of cutoff distances for association sites for each site type
            in each group type
        rd_klab : numpy.ndarray, Optional, default=None
            Position of association site in each group (nbead, nbead, nsite, nsite)
        reduction_ratio : float, Optional, default=0.25
            Reduced distance of the sites from the center of the sphere of
            interaction. This value is used when site position, rd_klab is None

        Returns
        -------
        Kijklab : numpy.ndarray
            Bonding volume for each molecule and site combination.
        """

        self._check_temperature_dependent_parameters(T)

        dij_bar = np.zeros((self.ncomp, self.ncomp))
        for i in range(self.ncomp):
            for j in range(self.ncomp):
                dij_bar[i, j] = np.mean([self.eos_dict["dii_eff"][i], self.eos_dict["dii_eff"][j]])

        Kijklab = Aassoc.calc_bonding_volume(rc_klab, dij_bar, rd_klab=rd_klab, reduction_ratio=reduction_ratio)

        return Kijklab

    def parameter_refresh(self, bead_library, cross_library):
        r"""
        To refresh dependent parameters

        Those parameters that are dependent on bead_library and cross_library
        attributes **must** be updated by running this function after all parameters
        from update_parameters method have been changed.

        Attributes
        ----------
        alphakl : np.array
            (Ngroup,Ngroup) "A dimensionless form of the integrated vdW energy of the
            Mie potential" eq. 33
        eos_dict : dict
            The following entries are updated:

            - Ckl (numpy.ndarray) - Matrix of Mie potential prefactors between
            beads (l,k)
            - epsilonkl (numpy.ndarray) - Matrix of Mie potential well depths for
            groups (k,l)
            - sigmakl (numpy.ndarray) - Matrix of bead diameters (k,l)
            - lambdarkl (numpy.ndarray) - Matrix of repulsive Mie exponent for
            groups (k,l)
            - lambdaakl (numpy.ndarray) - Matrix of attractive Mie exponent for
            groups (k,l)
            - Cmol2seg (float) - Conversion factor from from molecular number density,
            :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.
            - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
            group) k multiplied by that of bead l

        """

        self.bead_library.update(bead_library)
        self.cross_library.update(cross_library)

        self.eos_dict["Sk"] = tb.extract_property("Sk", self.bead_library, self.beads, default=1.0)

        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigmakl"] = output["sigma"]
        self.eos_dict["epsilonkl"] = output["epsilon"]
        self.eos_dict["lambdaakl"] = output["lambdaa"]
        self.eos_dict["lambdarkl"] = output["lambdar"]

        # Update Non bonded matrices
        if not np.isnan(self.T) and self.T is not None:
            self._check_temperature_dependent_parameters(self.T)
        else:
            self._check_temperature_dependent_parameters(298)

        # Initiate average interaction terms
        self.calc_component_averaged_properties()

        if not np.any(np.isnan(self.xi)):
            self.eos_dict["Cmol2seg"], self.eos_dict["xskl"] = stb.calc_composition_dependent_variables(
                self.xi,
                self.eos_dict["molecular_composition"],
                self.bead_library,
                self.beads,
            )

        self.eos_dict["Ckl"] = prefactor(self.eos_dict["lambdarkl"], self.eos_dict["lambdaakl"])
        self.eos_dict["alphakl"] = self.eos_dict["Ckl"] * (
            (1.0 / (self.eos_dict["lambdaakl"] - 3.0)) - (1.0 / (self.eos_dict["lambdarkl"] - 3.0))
        )

    def _check_density(self, rho):
        r"""
        This function checks that the density array is in the correct format for
        further calculations.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]

        Returns
        -------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]

        """

        if np.isscalar(rho):
            rho = np.array([rho])
        elif not isinstance(rho, np.ndarray):
            rho = np.array(rho)
        if len(np.shape(rho)) == 2:
            rho = rho[0]

        if any(np.isnan(rho)):
            raise ValueError("NaN was given as a value of density, rho")
        elif rho.size == 0:
            raise ValueError("No value of density, rho, was given")
        elif any(rho < 0.0):
            raise ValueError("Density values cannot be negative.")

        return rho

    def _check_temperature_dependent_parameters(self, T):
        r"""
        This function checks that the temperature dependent parameters are computed
        for the correct value. If not, they are recomputed.

        Parameters
        ----------
        T : float
            Temperature of the system [K]

        Attributes
        ---------
        T : float
            Updated temperature value
        eos_dict : dict
            The following entries are updated:

            - dkl (numpy.ndarray) - Matrix of hard sphere equivalent for each bead and
            interaction between them (l,k)
            - x0kl (numpy.ndarray) - Matrix of sigmakl/dkl, sigmakl is the Mie radius
            for groups (k,l)
            - etc. Other matrices will also be updated if temperature dependent
            multipole mixing rules are used.

        """

        if self.T != T:
            self.T = T
            # Check for temperature dependent mixing rule
            if self._mixing_temp_dependence is None:
                self._mixing_temp_dependence = False
                for key, value in self.combining_rules.items():
                    if "temperature" in value:
                        self._mixing_temp_dependence = True
                        if "additional_outputs" in value:
                            for params in value["additional_outputs"]:
                                self.combining_rules[params]["function"] = "None"
                        self.combining_rules[key]["temperature"] = T
            else:
                for key, value in self.combining_rules.items():
                    if "temperature" in value:
                        self.combining_rules[key]["temperature"] = T

            if self._mixing_temp_dependence:
                output = tb.cross_interaction_from_dict(
                    self.beads,
                    self.bead_library,
                    self.combining_rules,
                    cross_library=self.cross_library,
                )
                self.eos_dict["sigmakl"] = output["sigma"]
                self.eos_dict["epsilonkl"] = output["epsilon"]
                self.eos_dict["lambdaakl"] = output["lambdaa"]
                self.eos_dict["lambdarkl"] = output["lambdar"]

                # compute alphakl eq. 33
                self.eos_dict["Ckl"] = prefactor(self.eos_dict["lambdarkl"], self.eos_dict["lambdaakl"])
                self.eos_dict["alphakl"] = self.eos_dict["Ckl"] * (
                    (1.0 / (self.eos_dict["lambdaakl"] - 3.0)) - (1.0 / (self.eos_dict["lambdarkl"] - 3.0))
                )
                self.calc_component_averaged_properties()

            self.eos_dict["dkl"], self.eos_dict["x0kl"] = stb.calc_hard_sphere_matricies(
                T,
                self.eos_dict["sigmakl"],
                self.bead_library,
                self.beads,
                prefactor,
            )
            self._update_chain_temperature_dependent_variables(T)

    def _check_composition_dependent_parameters(self, xi):
        r"""
        This function checks that the composition dependent parameters are computed
        for the correct value. If not, they are recomputed.

        Parameters
        ----------
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Attributes
        ---------
        xi : numpy.ndarray
            Component mole fractions are updated
        eos_dict : dict
            The following entries are updated:

            - Cmol2seg (float) - Conversion factor from from molecular number density,
            :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.
            - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
            group) k multiplied by that of bead l

        """
        xi = np.array(xi)
        if not np.all(self.xi == xi):
            self.eos_dict["Cmol2seg"], self.eos_dict["xskl"] = stb.calc_composition_dependent_variables(
                xi,
                self.eos_dict["molecular_composition"],
                self.bead_library,
                self.beads,
            )
            self.xi = xi

    def _update_chain_temperature_dependent_variables(self, T):
        r"""
        This function checks that the temperature dependent parameters for the chain
        contribution are computed for the correct value. If not, they are recomputed.

        Parameters
        ----------
        T : float
            Temperature of the system [K]

        Attributes
        ---------
        T : float
            Updated temperature value
        eos_dict : dict
            The following entries are updated:

            - dii_eff (numpy.ndarray) - Matrix of mole averaged hard sphere equivalent
            for each bead and interaction between them (i.j)
            - x0ii (numpy.ndarray) - Matrix of sigmaii_avg/dii_eff, sigmaii_avg is the
            average molecular Mie radius and dii_eff the average molecular hard sphere
            diameter

        """

        zki = np.zeros((self.ncomp, self.nbeads), float)
        zkinorm = np.zeros(self.ncomp, float)
        dii_eff = np.zeros((self.ncomp), float)
        # compute zki
        for i in range(self.ncomp):
            for k in range(self.nbeads):
                zki[i, k] = (
                    self.eos_dict["molecular_composition"][i, k] * self.eos_dict["Vks"][k] * self.eos_dict["Sk"][k]
                )
                zkinorm[i] += zki[i, k]

        for i in range(self.ncomp):
            for k in range(self.nbeads):
                zki[i, k] = zki[i, k] / zkinorm[i]

        for i in range(self.ncomp):
            for k in range(self.nbeads):
                for l in range(self.nbeads):
                    dii_eff[i] += zki[i, k] * zki[i, l] * self.eos_dict["dkl"][k, l] ** 3
            dii_eff[i] = dii_eff[i] ** (1 / 3.0)
        self.eos_dict["dii_eff"] = dii_eff

        # compute x0ii
        self.eos_dict["x0ii"] = self.eos_dict["sigmaii_avg"] / dii_eff

    def __str__(self):

        string = "Beads: {}".format(self.beads)
        return string



"""
DESPASITO
DESPASITO: Determining Equilibrium State and Parametrization Application for SAFT,
Intended for Thermodynamic Output
"""

import os
import logging
import logging.handlers
from ._version import __version__ as __version__

logger = logging.getLogger("despasito")
logger.setLevel(30)


def initiate_logger(console=None, log_file=None, verbose=30):
    """
    Initiate a logging handler if more detail on the calculations is desired.

    This function is useful when DESPASITO is used as an imported package. If a handler
    of the given type is already present, nothing is done, as is the case when the
    input file schema is used. If either handler is given a value of False, the handler
    of that type is removed.

    Parameters
    ----------
    console : bool, Optional, default=None
        Initiates a stream handler to print to a console. If True, this handler is
        initiated. If it is False, then any StreamHandler is removed.
    log_file : bool/str, Optional, default=None
        If log output should be recorded in a file, set this keyword to either True or
        to a name for the log file. If True, the file name 'despasito.log' is used.
        Note that if a file with the same name already exists, it will be deleted.
    verbose : int, Optional, default=30
        The verbosity of logging information can be set to any supported representation
        of the
        `logging level <LOGGER>`_.

    .. _LOGGER: https://docs.python.org/3/library/logging.html#logging-levels
    """

    logger.setLevel(verbose)

    # Check for existing handlers
    handler_console = None
    handler_logfile = None
    for tmp in logger.handlers:
        if "RotatingFileHandler" in str(tmp):
            handler_logfile = tmp
        if "StreamHandler" in str(tmp):
            handler_console = tmp

    # Set up logging to console
    if console and handler_console is None:
        console_handler = logging.StreamHandler()  # sys.stderr
        console_handler.setFormatter(logging.Formatter("[%(levelname)s](%(name)s): %(message)s"))
        console_handler.setLevel(verbose)
        logger.addHandler(console_handler)
    elif console:
        logger.warning("StreamHandler already exists")
    elif handler_console is not None:
        handler_console.close()
        logger.removeHandler(handler_console)

    # Rotating File Handler
    if log_file is not None and handler_logfile is None:

        if not isinstance(log_file, str):
            log_file = "despasito.log"

        if os.path.isfile(log_file):
            os.remove(log_file)

        log_file_handler = logging.handlers.RotatingFileHandler(log_file)
        log_file_handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s](%(name)s:%(funcName)s:%(lineno)d): " + "%(message)s")
        )
        log_file_handler.setLevel(verbose)
        logger.addHandler(log_file_handler)
    elif log_file:
        logger.warning("RotatingFileHandler already exists")
    elif not handler_logfile and handler_logfile is not None:
        handler_logfile.close()
        logger.removeHandler(handler_logfile)



import numpy as np

import despasito
import despasito.input_output.read_input as io
import despasito.parameter_fitting as fit
from despasito.equations_of_state import initiate_eos

despasito.initiate_logger(console=True, verbose=10)

Eos = initiate_eos(
    eos="saft.gamma_mie",
    beads=["CH2", "CH3"],
    molecular_composition=np.array([[2.0, 2.0]]),
    bead_library=io.json_to_dict("../../library/SAFTgroup.json"),
)

fit.fit(
    optimization_parameters={
        "fit_bead": "CH3",
        "fit_parameter_names": ["epsilon_CH2"],
        "epsilon_CH2_bounds": [150.0, 600.0],
        "parameters_guess": [300.0],
    },
    exp_data={
        "Knovel": {
            "data_class_type": "liquid_density",
            "eos_obj": Eos,
            "calculation_type": "liquid_properties",
            "T": np.array([272.15, 323.15, 298.15]),
            "rhol": np.array([10357.0, 10364.8, 10140.0]),
            "delta": np.array([14453.0, 13700.0, 14100.0]),
        },
    },
    global_opts={"method": "grid_minimization", "Ns": 3},
)



import numpy as np

import despasito
import despasito.input_output.read_input as io
import despasito.thermodynamics as thermo
from despasito.equations_of_state import initiate_eos

despasito.initiate_logger(console=True, verbose=10)

Eos = initiate_eos(
    eos="saft.gamma_mie",
    beads=["CH3", "CH2"],
    molecular_composition=np.array([[2.0, 4.0], [2.0, 5.0]]),
    bead_library=io.json_to_dict("../../library/SAFTgroup.json"),
    cross_library=io.json_to_dict("../../library/SAFTcross.json"),
)

output = thermo.thermo(
    Eos,
    calculation_type="liquid_properties",
    Tlist=320.0,
    Plist=1e5,
    xilist=np.array([0.4, 0.6]),
)

print("Thermo Output", output)
args = (output["rhol"][0], 320.0, [0.4, 0.6])
print("Helmholtz Contributions:")
print("    Ideal: ", Eos.Aideal(*args))
print("    Monomer: ", Eos.saft_source.Amonomer(*args))
print("    Chain: ", Eos.saft_source.Achain(*args))
